"""
The decaydata module defines the ``DecayData``, ``DecayMatricesScipy`` and ``DecayMatricesSympy``
classes. ``DecayMatricesScipy`` and ``DecayMatricesSympy`` instances store data used for decay
calculations and other inventory transformations (using double-precision floating point and SymPy
arbitrary precision numbers, respectively). ``DecayData`` instances store ``DecayMatricesScipy``,
``DecayMatricesSympy`` and unit converter objects, along with other decay data for reporting to
users (this data is not used for calculations).

The docstring code examples assume that ``radioactivedecay`` has been imported
as ``rd``:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

Constants
---------
DEFAULTDATA : DecayData
    Default radioactive decay dataset used by ``radioactivedecay``. This is currently ICRP-107 for
    decay data and AME2020 and Nubase2020 for atomic mass data.
"""

import pathlib
import pickle
import platform
from abc import ABC, abstractmethod
from typing import Any, Optional, Union

import numpy as np
import pkg_resources
import sympy
from scipy import sparse
from sympy import Matrix
from sympy.core.expr import Expr
from sympy.matrices import SparseMatrix

from radioactivedecay.converters import UnitConverterFloat
from radioactivedecay.utils import parse_nuclide

# importlib.resources API from Python version 3.9+ is used
# this block adds support using a backport for Python <3.9
# the importlib_resources dependency can be removed once min required Python version is 3.9
if pkg_resources.parse_version(
    platform.python_version()
) >= pkg_resources.parse_version("3.9"):
    from importlib import resources
else:
    import importlib_resources as resources  # type:ignore


def _csr_matrix_equal(matrix_a: sparse.csr_matrix, matrix_b: sparse.csr_matrix) -> bool:
    """
    Utility function checking whether two SciPy Compressed Sparse Row (CSR) matrices are equal
    (i.e. they have the same elements and all elements are equal).

    Parameters
    ----------
    matrix_a : scipy.sparse.csr.csr_matrix
        First SciPy CSR matrix.
    matrix_b : scipy.sparse.csr.csr_matrix
        Second SciPy CSR matrix.

    Returns
    -------
    bool
        True if both sparse matrices are equal, False otherwise.
    """

    return (
        np.array_equal(matrix_a.indptr, matrix_b.indptr)
        and np.array_equal(matrix_a.indices, matrix_b.indices)
        and np.array_equal(matrix_a.data, matrix_b.data)
    )


class DecayMatrices(ABC):
    """
    Template class for for SciPy (float) or SymPy DecayMatrices classes. Instances of these classes
    store vectors and matrices of data used for decay calculations.

    Parameters
    ----------
    atomic_masses : numpy.ndarray or sympy.matrices.dense.MutableSparseMatrix
        Column vector of the atomic masses (in g/mol).
    decay_consts : numpy.ndarray or sympy.matrices.dense.MutableSparseMatrix
        Column vector of the decay constants (in s\\ :sup:`-1`).
    matrix_c : scipy.sparse.csr.csr_matrix or sympy.matrices.dense.MutableSparseMatrix
        A pre-calculated sparse lower traingular matrix used in decay calculations.
    matrix_c_inv : scipy.sparse.csr.csr_matrix or sympy.matrices.dense.MutableSparseMatrix
        The inverse of matrix_c, also used in decay calculations.

    Attributes
    ----------
    atomic_masses : numpy.ndarray or sympy.matrices.dense.MutableSparseMatrix
        Column vector of the atomic masses (in g/mol).
    decay_consts : numpy.ndarray or sympy.matrices.dense.MutableSparseMatrix
        Column vector of the decay constants (in s\\ :sup:`-1`).
    ln2: float or sympy.log
        Constant natural logarithm of 2.
    matrix_c : scipy.sparse.csr.csr_matrix or sympy.matrices.dense.MutableSparseMatrix
        A precalculated sparse lower triangular matrix used in decay calculations.
    matrix_c_inv : scipy.sparse.csr.csr_matrix or sympy.matrices.dense.MutableSparseMatrix
        The inverse of matrix_c, also used in decay calculations.
    matrix_e : scipy.sparse.csr.csr_matrix or sympy.matrices.dense.MutableSparseMatrix
        The matrix exponential that is used in radioactive decay calculations. It is a diagonal
        matrix that is pre-allocted for performance reasons.
    vector_n0 : numpy.ndarray or sympy.matrices.dense.MutableDenseMatrix
        Column vector for the number of atoms of each nuclide. It is pre-allocted here to improve
        the performance of decay calculations.

    """

    def __init__(
        self,
        atomic_masses: Union[np.ndarray, Matrix],
        decay_consts: Union[np.ndarray, Matrix],
        matrix_c: Union[sparse.csr_matrix, Matrix],
        matrix_c_inv: Union[sparse.csr_matrix, Matrix],
    ) -> None:
        self.atomic_masses = atomic_masses
        self.decay_consts = decay_consts
        self.ln2 = self._setup_ln2()
        self.matrix_c = matrix_c
        self.matrix_c_inv = matrix_c_inv
        self.matrix_e = self._setup_matrix_e(matrix_c.shape[0])
        self.vector_n0 = self._setup_vector_n0(matrix_c.shape[0])

    @staticmethod
    @abstractmethod
    def _setup_ln2() -> float:
        """
        Set up natural log of 2 (ln(2)).
        """

    @staticmethod
    @abstractmethod
    def _setup_matrix_e(size: int) -> Union[sparse.csr_matrix, Matrix]:
        """
        Set up sparse matrix_e (created here rather than at decay calculation time
        for performance reasons).
        """

    @staticmethod
    @abstractmethod
    def _setup_vector_n0(size: int) -> Union[np.ndarray, Matrix]:
        """
        Set up vector_n0 (created here rather than at decay calculation time for performance
        reasons).
        """

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``DecayMatrices`` instances are equal with ``==`` operator.
        """

    def __ne__(self, other: object) -> bool:
        """
        Check whether two ``DecayMatrices`` instances are not equal with ``!=`` operator.
        """

        if not isinstance(other, self.__class__):
            return NotImplemented
        return not self.__eq__(other)


class DecayMatricesScipy(DecayMatrices):
    """
    Version of DecayMatrices with data for SciPy double-precision floating point decay
    calculations.

    Parameters
    ----------
    atomic_masses : numpy.ndarray
        Column vector of the atomic masses (in g/mol).
    decay_consts : numpy.ndarray
        Column vector of the decay constants (in s\\ :sup:`-1`).
    matrix_c : scipy.sparse.csr.csr_matrix
        A pre-calculated sparse lower traingular matrix used in decay calculations.
    matrix_c_inv : scipy.sparse.csr.csr_matrix
        The inverse of matrix_c, also used in decay calculations.

    Attributes
    ----------
    atomic_masses : numpy.ndarray
        Column vector of the atomic masses (in g/mol).
    decay_consts : numpy.ndarray
        Column vector of the decay constants (in s\\ :sup:`-1`).
    ln2: float
        Constant natural logarithm of 2.
    matrix_c : scipy.sparse.csr.csr_matrix
        A precalculated sparse lower triangular matrix used in decay calculations.
    matrix_c_inv : scipy.sparse.csr.csr_matrix
        The inverse of matrix_c, also used in decay calculations.
    matrix_e : scipy.sparse.csr.csr_matrix
        The matrix exponential that is used in radioactive decay calculations. It is a diagonal
        matrix that is pre-allocted for performance reasons.
    vector_n0 : numpy.ndarray
        Column vector for the number of atoms of each nuclide. It is pre-allocted here to improve
        the performance of decay calculations.

    """

    @staticmethod
    def _setup_ln2() -> float:
        """
        Set up natural log of 2 (ln(2)).
        """

        log2: float = np.log(2)
        return log2

    @staticmethod
    def _setup_matrix_e(size: int) -> sparse.csr_matrix:
        """
        Set up sparse matrix_e (created here rather than at decay calculation time
        for performance reasons).
        """

        return sparse.csr_matrix(
            (
                np.zeros(size),
                (np.arange(size), np.arange(size)),
            )
        )

    @staticmethod
    def _setup_vector_n0(size: int) -> np.ndarray:
        """
        Set up vector_n0 (created here rather than at decay calculation time for performance
        reasons).
        """

        return np.zeros([size], dtype=np.float64)

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``DecayMatricesScipy`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            np.array_equal(self.atomic_masses, other.atomic_masses)
            and np.array_equal(self.decay_consts, other.decay_consts)
            and self.ln2 == other.ln2
            and _csr_matrix_equal(self.matrix_c, other.matrix_c)
            and _csr_matrix_equal(self.matrix_c_inv, other.matrix_c_inv)
            and _csr_matrix_equal(self.matrix_e, other.matrix_e)
            and np.array_equal(self.vector_n0, other.vector_n0)
        )

    def __repr__(self) -> str:
        return (
            "DecayMatricesScipy: data stored in SciPy/NumPy objects for double precision "
            "calculations."
        )


class DecayMatricesSympy(DecayMatrices):
    """
    Version of DecayMatrices with data for SymPy arbitrary precision decay calculations.

    Parameters
    ----------
    atomic_masses : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the atomic masses (in g/mol).
    decay_consts : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the decay constants (in s\\ :sup:`-1`).
    matrix_c : sympy.matrices.sparse.MutableSparseMatrix
        A pre-calculated sparse lower traingular matrix used in decay calculations.
    matrix_c_inv : sympy.matrices.sparse.MutableSparseMatrix
        The inverse of matrix_c, also used in decay calculations.

    Attributes
    ----------
    atomic_masses : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the atomic masses (in g/mol).
    decay_consts : sympy.matrices.dense.MutableDenseMatrix
        Column vector of the decay constants (in s\\ :sup:`-1`).
    ln2: sympy.log
        Constant natural logarithm of 2.
    matrix_c : sympy.matrices.sparse.MutableSparseMatrix
        A precalculated sparse lower triangular matrix used in decay calculations.
    matrix_c_inv : sympy.matrices.sparse.MutableSparseMatrix
        The inverse of matrix_c, also used in decay calculations.
    matrix_e : sympy.matrices.sparse.MutableSparseMatrix
        The matrix exponential that is used in radioactive decay calculations. It is a diagonal
        matrix that is pre-allocted for performance reasons.
    vector_n0 : sympy.matrices.dense.MutableDenseMatrix
        Column vector for the number of atoms of each nuclide. It is pre-allocted here to improve
        the performance of decay calculations.

    """

    @staticmethod
    def _setup_ln2() -> sympy.log:
        """
        Returns SymPy version of ln2.
        """

        return sympy.log(2)

    @staticmethod
    def _setup_matrix_e(size: int) -> Matrix:
        """
        Returns template for SymPy version of matrix_e.
        """

        return SparseMatrix.zeros(size, size)

    @staticmethod
    def _setup_vector_n0(size: int) -> Matrix:
        """
        Returns template for SymPy version of vector_n0.
        """

        return Matrix.zeros(size, 1)

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``DecayMatricesSympy`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, DecayMatricesSympy):
            return NotImplemented
        return (
            self.atomic_masses == other.atomic_masses
            and self.decay_consts == other.decay_consts
            and self.ln2 == other.ln2
            and self.matrix_c == other.matrix_c
            and self.matrix_c_inv == other.matrix_c_inv
            and self.matrix_e == other.matrix_e
            and self.vector_n0 == other.vector_n0
        )

    def __repr__(self) -> str:
        return (
            "DecayMatricesSympy: data stored in SymPy objects for arbitrary-precision "
            "calculations."
        )


class DecayData:
    """
    Instances of DecayData store a complete radioactive decay dataset. It stores data for decay
    calculations and inventory transformations using DecayMatrices attributes. Other decay data
    which are reported to users but not used incalculations are stored in other instance
    attributes.

    Parameters
    ----------
    dataset_name : str
        Name of the decay dataset.
    bfs : numpy.ndarray
        NumPy array of lists with branching fraction data.
    float_year_conv : float
        Number of days in one year for float time unit conversions.
    hldata : numpy.ndarray
        List of tuples containing half-life floats, time unit strings and readable format half-life
        strings.
    modes : numpy.ndarray
        NumPy array of lists with decay mode data.
    nuclides : numpy.ndarray
        NumPy array of nuclides in the dataset (string format is 'H-3', etc.).
    progeny : numpy.ndarray
        NumPy array of lists with direct progeny data.
    scipy_data : DecayMatricesScipy
        Dataset of double precision decay matrices (SciPy/NumPy objects).
    sympy_data : None or DecayMatricesSympy
        Dataset of arbitrary-precision decay matrices (SymPy objects). Or None if SymPy
        functionality is not used. Default is None.
    sympy_year_conv : None or Expr
        Number of days in one year for SymPy time unit conversions.

    Attributes
    ----------
    bfs : numpy.ndarray
        NumPy array of lists with branching fraction data.
    dataset_name : str
        Name of the decay dataset.
    float_year_conv : float
        Number of days in one year for float time unit conversions.
    hldata : numpy.ndarray
        List of tuples containing half-life floats, time unit strings and readable format half-life
        strings.
    modes : numpy.ndarray
        NumPy array of lists with decay mode data.
    nuclides : numpy.ndarray
        NumPy array of nuclides in the dataset (string format is 'H-3', etc.).
    nuclide_dict : dict
        Dictionary containing nuclide strings as keys and positions in the matrices as values.
    progeny : numpy.ndarray
        NumPy array of lists with direct progeny data.
    _scipy_data : DecayMatricesScipy
        Dataset of double precision decay matrices (SciPy/NumPy objects).
    _sympy_year_conv : None or Expr
        Number of days in one year for SymPy time unit conversions.

    """

    def __init__(
        self,
        dataset_name: str,
        bfs: np.ndarray,
        float_year_conv: float,
        hldata: np.ndarray,
        modes: np.ndarray,
        nuclides: np.ndarray,
        progeny: np.ndarray,
        scipy_data: DecayMatricesScipy,
        sympy_data: Optional[DecayMatricesSympy] = None,
        sympy_year_conv: Optional[Expr] = None,
    ) -> None:
        self.dataset_name = dataset_name
        self.bfs = bfs
        self.float_year_conv = float_year_conv
        self.hldata = hldata
        self.modes = modes
        self.nuclides = nuclides
        self.nuclide_dict = dict(zip(self.nuclides, list(range(0, self.nuclides.size))))
        self.progeny = progeny
        self.scipy_data = scipy_data

        self._sympy_data: Optional[DecayMatricesSympy] = None
        self._sympy_year_conv: Optional[Expr] = None

        if sympy_data and sympy_year_conv:
            self._sympy_data = sympy_data
            self._sympy_year_conv = sympy_year_conv

    @property
    def sympy_data(self) -> DecayMatricesSympy:
        """
        Property to return DecayMatricesSympy instance if it exists, or raise an error if the
        dataset does not contain SymPy data.

        Raises
        ------
        ValueError
            If the DecayData instance does not contain any SymPy decay data.

        """

        if self._sympy_data is None:
            raise ValueError(f"{self.dataset_name} does not contain SymPy data.")
        return self._sympy_data

    @property
    def sympy_year_conv(self) -> Expr:
        """
        Property to return SymPy number of days in one year if it exists, or raise an error if the
        dataset does not contain SymPy data.

        Raises
        ------
        ValueError
            If the DecayData instance does not contain any SymPy decay data.

        """

        if self._sympy_year_conv is None:
            raise ValueError(f"{self.dataset_name} does not contain SymPy data.")
        return self._sympy_year_conv

    def half_life(self, nuclide: str, units: str = "s") -> Union[float, str]:
        """
        Returns the half-life of the nuclide as a float in your chosen units, or as
        a human-readable string with appropriate units.

        Parameters
        ----------
        nuclide : str
            Nuclide string.
        units : str, optional
            Units for half-life. Options are 'ps', 'ns', 'μs', 'us', 'ms', 's', 'm', 'h', 'd', 'y',
            'ky', 'My', 'By', 'Gy', 'Ty', 'Py', and common spelling variations. Default is 's', i.e.
            seconds. Use 'readable' to get a string of the half-life in human-readable units.

        Returns
        -------
        float or str
            Nuclide half-life.

        Examples
        --------
        >>> rd.DEFAULTDATA.half_life('Rn-222', 'd')
        3.8235
        >>> rd.DEFAULTDATA.half_life('H-3')
        388781329.30560005
        >>> rd.DEFAULTDATA.half_life('H-3', 'readable')
        '12.32 y'
        >>> rd.DEFAULTDATA.half_life('He-3')
        inf
        >>> rd.DEFAULTDATA.half_life('He-3', 'readable')
        'stable'

        """

        nuclide = parse_nuclide(nuclide, self.nuclides, self.dataset_name)
        half_life: float
        unit: str
        readable_str: str
        half_life, unit, readable_str = self.hldata[self.nuclide_dict[nuclide]]

        if units == "readable":
            return readable_str

        return (
            half_life
            if unit == units
            else UnitConverterFloat.time_unit_conv(
                half_life,
                units_from=unit,
                units_to=units,
                year_conv=self.float_year_conv,
            )
        )

    def branching_fraction(self, parent: str, progeny: str) -> float:
        """
        Returns the branching fraction for parent to progeny (if it exists).

        Parameters
        ----------
        parent : str
            Nuclide string of the parent.
        progeny : str
            Nuclide string of the progeny (can be stable or radioactive nuclide).

        Returns
        -------
        float
            Branching fraction (or zero if progeny is not actually a direct progeny of parent).

        Examples
        --------
        >>> rd.DEFAULTDATA.branching_fraction('K-40', 'Ca-40')
        0.8914

        """

        parent = parse_nuclide(parent, self.nuclides, self.dataset_name)
        progeny = parse_nuclide(progeny, self.nuclides, self.dataset_name)
        for idx, prog in enumerate(self.progeny[self.nuclide_dict[parent]]):
            if prog == progeny:
                branching_fraction: float = self.bfs[self.nuclide_dict[parent]][idx]
                return branching_fraction
        return 0.0

    def decay_mode(self, parent: str, progeny: str) -> str:
        """
        Returns the type of decay mode between parent and progeny (if one exists). Note: the decay
        mode string returned is not a list of all the different radiation types emitted during the
        decay process between parent and progeny. It is the label defined in the decay dataset to
        classify the decay type (e.g. '\u03b1', '\u03b2-' or 'IT').

        Parameters
        ----------
        parent : str
            Nuclide string of the parent.
        progeny : str
            Nuclide string of the progeny.

        Returns
        -------
        str
            Decay mode (or '' if progeny is not actually a direct progeny of the parent).

        Examples
        --------
        >>> rd.DEFAULTDATA.decay_mode('K-40', 'Ca-40')
        '\u03b2-'

        """

        parent = parse_nuclide(parent, self.nuclides, self.dataset_name)
        progeny = parse_nuclide(progeny, self.nuclides, self.dataset_name)
        for idx, prog in enumerate(self.progeny[self.nuclide_dict[parent]]):
            if prog == progeny:
                decay_mode: str = self.modes[self.nuclide_dict[parent]][idx]
                return decay_mode
        return ""

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``DecayData`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, DecayData):
            return NotImplemented
        return (
            self.dataset_name == other.dataset_name
            and self.float_year_conv == other.float_year_conv
            and np.array_equal(self.hldata, other.hldata)
            and np.array_equal(self.nuclides, other.nuclides)
            and self.nuclide_dict == other.nuclide_dict
            and np.array_equal(self.progeny, other.progeny)
            and np.array_equal(self.bfs, other.bfs)
            and np.array_equal(self.modes, other.modes)
            and self.scipy_data == other.scipy_data
            and self._sympy_data == other._sympy_data
            and self._sympy_year_conv == other._sympy_year_conv
        )

    def __ne__(self, other: object) -> bool:
        """
        Check whether two ``DecayData`` instances are not equal with ``!=`` operator.
        """

        if not isinstance(other, DecayData):
            return NotImplemented
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return (
            f"Decay dataset: {self.dataset_name}, contains SymPy data: "
            f"{self._sympy_data is not None}"
        )


def _get_package_filepath(subpackage_dir: str, filename: str) -> pathlib.Path:
    """
    Returns the path to a file which is bundled as a sub-package within the
    ``radioactivedecay`` package.

    Parameters
    ----------
    subpackage_dir : str
        Name of the sub-package directory.
    filename : str
        Name of the file.

    Returns
    -------
    pathlib.Path
        File path for the decay dataset file.

    """

    return resources.files(f"{__package__}.{subpackage_dir}").joinpath(filename)


def _get_filepath(
    dataset_name: str, dir_path: Optional[str], filename: str
) -> pathlib.Path:
    """
    Returns the path to a decay dataset file (located either within a sub-package of
    ``radioactivedecay`` or within a local system directory).

    Parameters
    ----------
    dataset_name : str
        Name of the decay dataset (also sub-package directory).
    dir_path : None or str
        Local directory containing dataset files, if not loading from a sub-package.
    filename : str
        Name of the file.

    Returns
    -------
    pathlib.Path
        File path for the decay dataset file.

    """

    if dir_path:
        return pathlib.Path(f"{dir_path}/{filename}")
    return _get_package_filepath(dataset_name, filename)


def _load_package_pickle_file(subpackage_dir: str, filename: str) -> Any:
    """
    Returns an object loaded from a pickle file which is bundled as a sub-package within the
    ``radioactivedecay`` package.

    Parameters
    ----------
    subpackage_dir : str
        Name of the sub-package directory.
    filename : str
        Name of the pickle file.

    Returns
    -------
    Object
        Object loaded from the pickle file.

    """

    with resources.files(f"{__package__}.{subpackage_dir}").joinpath(filename).open(
        "rb"
    ) as file:
        return pickle.load(file)


def _load_pickle_file(dataset_name: str, dir_path: Optional[str], filename: str) -> Any:
    """
    Load an object from a pickle file which is located either in a sub-package of
    ``radioactivedecay`` or within a local system directory.

    Parameters
    ----------
    dataset_name : str
        Name of the decay dataset (also sub-package directory).
    dir_path : None or str
        Local directory containing dataset files, if not loading from a sub-package.
    filename : str
        Name of the pickle file.

    Returns
    -------
    Object
        Object loaded from the pickle file.

    """

    if dir_path:
        with open(f"{dir_path}/{filename}", "rb") as file:
            obj = pickle.load(file)
        return obj
    return _load_package_pickle_file(dataset_name, filename)


def load_dataset(
    dataset_name: str, dir_path: Optional[str] = None, load_sympy: bool = False
) -> DecayData:
    """
    Load a decay dataset, either from a set of data files packaged within ``radioactivedecay``,
    or by specifying a local directory containing the data files.

    Parameters
    ----------
    dataset_name : str
        Name of the decay dataset (or sub-package directory name).
    dir_path : None or str, optional
        Path to the directory containing the decay dataset files. Use None to load data that are
        bundled as a sub-package of ``radioactivedecay`` (dataset_name should match the
        sub-package name in this case). Default is None.
    load_sympy : bool, optional
        Load SymPy version of the decay data for arbitrary-precision decay calculations. Default is
        False.

    Returns
    -------
    DecayData
        A decay dataset used by ``radioactivedecay``.

    """

    data = np.load(
        _get_filepath(dataset_name, dir_path, "decay_data.npz"),
        allow_pickle=True,
    )

    decay_consts: np.ndarray = np.array(
        [
            np.log(2)
            / UnitConverterFloat.time_unit_conv(
                hl[0], units_from=hl[1], units_to="s", year_conv=data["year_conv"]
            )
            for hl in data["hldata"]
        ]
    )

    matrix_c = sparse.load_npz(_get_filepath(dataset_name, dir_path, "c_scipy.npz"))
    matrix_c_inv = sparse.load_npz(
        _get_filepath(dataset_name, dir_path, "c_inv_scipy.npz")
    )

    scipy_data = DecayMatricesScipy(
        data["masses"], decay_consts, matrix_c, matrix_c_inv
    )

    sympy_data = None
    sympy_year_conv = None
    if load_sympy:
        sympy_pickle_version = (
            "1.8"
            if pkg_resources.parse_version(sympy.__version__)
            < pkg_resources.parse_version("1.9")
            else "1.9"
        )

        atomic_masses = _load_pickle_file(
            dataset_name, dir_path, f"atomic_masses_sympy_{sympy_pickle_version}.pickle"
        )
        decay_consts = _load_pickle_file(
            dataset_name, dir_path, f"decay_consts_sympy_{sympy_pickle_version}.pickle"
        )
        matrix_c = _load_pickle_file(
            dataset_name, dir_path, f"c_sympy_{sympy_pickle_version}.pickle"
        )
        matrix_c_inv = _load_pickle_file(
            dataset_name, dir_path, f"c_inv_sympy_{sympy_pickle_version}.pickle"
        )
        sympy_year_conv = _load_pickle_file(
            dataset_name,
            dir_path,
            f"year_conversion_sympy_{sympy_pickle_version}.pickle",
        )
        sympy_data = DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )

    return DecayData(
        dataset_name,
        data["bfs"],
        data["year_conv"],
        data["hldata"],
        data["modes"],
        data["nuclides"],
        data["progeny"],
        scipy_data,
        sympy_data,
        sympy_year_conv,
    )


DEFAULTDATA = load_dataset("icrp107_ame2020_nubase2020", load_sympy=True)



"""
Unit tests for decaydata.py functions, classes and methods.
"""

import unittest

import numpy as np
from scipy import sparse
from sympy import Integer, Matrix, log
from sympy.matrices import SparseMatrix

from radioactivedecay import decaydata, icrp107_ame2020_nubase2020


class TestFunctions(unittest.TestCase):
    """
    Unit tests for the decaydata.py functions.
    """

    # pylint: disable=protected-access

    def test__csr_matrix_equal(self) -> None:
        """
        Test function to check equality of two SciPy Compressed Sparse Row (CSR) matrices.
        """

        matrix_a = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_b = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c = sparse.csr_matrix(([1.0], ([1], [0])), shape=(2, 2))
        self.assertEqual(decaydata._csr_matrix_equal(matrix_a, matrix_b), True)
        self.assertEqual(decaydata._csr_matrix_equal(matrix_a, matrix_c), False)


class TestDecayMatricesScipy(unittest.TestCase):
    """
    Unit tests for the decaydata.py DecayMatricesScipy class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of DecayMatricesScipy objects.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([1], [1])), shape=(2, 2))
        decay_mats = decaydata.DecayMatricesScipy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats.atomic_masses[0], 0.0)
        self.assertEqual(decay_mats.atomic_masses[1], 0.0)
        self.assertEqual(decay_mats.decay_consts[0], 0.0)
        self.assertEqual(decay_mats.decay_consts[1], 0.0)
        self.assertEqual(decay_mats.ln2, np.log(2))
        self.assertEqual(decay_mats.matrix_c[0, 0], 1.0)
        self.assertEqual(decay_mats.matrix_c_inv[1, 1], 1.0)
        self.assertEqual(decay_mats.matrix_e[0, 0], 0.0)
        self.assertEqual(decay_mats.matrix_e[1, 1], 0.0)
        self.assertEqual(decay_mats.vector_n0[0], 0.0)
        self.assertEqual(decay_mats.vector_n0[1], 0.0)

    def test___eq__(self) -> None:
        """
        Test DecayMatricesScipy equality.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        decay_mats_a = decaydata.DecayMatricesScipy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        decay_mats_b = decaydata.DecayMatricesScipy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats_a, decay_mats_b)

        self.assertFalse(decay_mats_a == "random object")

    def test___ne__(self) -> None:
        """
        Test DecayMatricesScipy inequal.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        decay_mats_a = decaydata.DecayMatricesScipy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([2.0], ([0], [0])), shape=(2, 2))
        decay_mats_b = decaydata.DecayMatricesScipy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertNotEqual(decay_mats_a, decay_mats_b)

        self.assertTrue(decay_mats_a != "random object")

    def test___repr__(self) -> None:
        """
        Test DecayMatricesScipy __repr__ strings.
        """

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([1], [1])), shape=(2, 2))
        decay_mats = decaydata.DecayMatricesScipy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(
            repr(decay_mats),
            "DecayMatricesScipy: data stored in SciPy/NumPy objects for double precision "
            "calculations.",
        )


class TestDecayMatricesSympy(unittest.TestCase):
    """
    Unit tests for the decaydata.py DecayMatricesSympy class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of DecayMatricesSympy objects.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats.atomic_masses[0], 0.0)
        self.assertEqual(decay_mats.atomic_masses[1], 0.0)
        self.assertEqual(decay_mats.decay_consts[0], 0.0)
        self.assertEqual(decay_mats.decay_consts[1], 0.0)
        self.assertEqual(decay_mats.ln2, log(2))
        self.assertEqual(decay_mats.matrix_c[0, 0], Integer(2))
        self.assertEqual(decay_mats.matrix_c_inv[1, 1], Integer(3))
        self.assertEqual(decay_mats.matrix_e[0, 0], Integer(0))
        self.assertEqual(decay_mats.matrix_e[1, 1], Integer(0))
        self.assertEqual(decay_mats.vector_n0[0], 0.0)
        self.assertEqual(decay_mats.vector_n0[1], 0.0)

    def test___eq__(self) -> None:
        """
        Test DecayMatricesSympy equality.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats_a = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats_b = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(decay_mats_a, decay_mats_b)

        self.assertFalse(decay_mats_a == "random object")

    def test___ne__(self) -> None:
        """
        Test DecayMatricesSympy inequality.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats_a = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(5)
        decay_mats_b = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertNotEqual(decay_mats_a, decay_mats_b)

        self.assertTrue(decay_mats_a != "random object")

    def test___repr__(self) -> None:
        """
        Test DecayMatricesSympy __repr__ strings.
        """

        atomic_masses = Matrix.zeros(2, 1)
        decay_consts = Matrix.zeros(2, 1)
        matrix_c = SparseMatrix.zeros(2, 2)
        matrix_c[0, 0] = Integer(2)
        matrix_c_inv = SparseMatrix.zeros(2, 2)
        matrix_c_inv[1, 1] = Integer(3)
        decay_mats = decaydata.DecayMatricesSympy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )
        self.assertEqual(
            repr(decay_mats),
            "DecayMatricesSympy: data stored in SymPy objects for arbitrary-precision "
            + "calculations.",
        )


class TestDecayData(unittest.TestCase):
    """
    Unit tests for the decaydata.py DecayData class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of DecayData objects.
        """

        dataset_name = "test_dataset"

        bfs = np.array([[1.0], []], dtype=object)
        hldata = np.array([[2.0, "h", "2.0 h"], [np.inf, "s", "stable"]], dtype=object)
        modes = np.array([["\u03b2+"], []], dtype=object)
        nuclides = np.array(["A-2", "B-1"])
        progeny = np.array([["B-1"], []], dtype=object)
        float_year_conv = 365.2422

        atomic_masses = np.array([0.0] * 2)
        decay_consts = np.array([0.0] * 2)
        matrix_c = sparse.csr_matrix(([1.0], ([0], [0])), shape=(2, 2))
        matrix_c_inv = sparse.csr_matrix(([1.0], ([1], [1])), shape=(2, 2))
        decay_mats = decaydata.DecayMatricesScipy(
            atomic_masses, decay_consts, matrix_c, matrix_c_inv
        )

        dataset = decaydata.DecayData(
            dataset_name,
            bfs,
            float_year_conv,
            hldata,
            modes,
            nuclides,
            progeny,
            decay_mats,
        )

        self.assertEqual(dataset.dataset_name, "test_dataset")
        self.assertEqual(dataset.hldata[0][0], 2.0)
        self.assertEqual(dataset.hldata[0][1], "h")
        self.assertEqual(dataset.hldata[0][2], "2.0 h")
        self.assertEqual(dataset.hldata[-1][0], np.inf)
        self.assertEqual(dataset.hldata[-1][1], "s")
        self.assertEqual(dataset.hldata[-1][2], "stable")
        self.assertEqual(dataset.nuclides[0], "A-2")
        self.assertEqual(dataset.nuclides[-1], "B-1")
        self.assertEqual(dataset.nuclide_dict["A-2"], 0)
        self.assertEqual(dataset.nuclide_dict["B-1"], 1)
        self.assertEqual(dataset.progeny[0][0], "B-1")
        self.assertEqual(dataset.bfs[0][0], 1.0)
        self.assertEqual(dataset.modes[0][0], "\u03b2+")
        self.assertEqual(dataset.progeny[-1], [])
        self.assertEqual(dataset.bfs[-1], [])
        self.assertEqual(dataset.modes[-1], [])
        self.assertEqual(dataset.float_year_conv, 365.2422)
        with self.assertRaises(ValueError):
            assert dataset.sympy_data
        with self.assertRaises(ValueError):
            assert dataset.sympy_year_conv

        atomic_masses_sympy = Matrix.zeros(2, 1)
        decay_consts_sympy = Matrix.zeros(2, 1)
        matrix_c_sympy = SparseMatrix.zeros(2, 2)
        matrix_c_sympy[0, 0] = Integer(2)
        matrix_c_inv_sympy = SparseMatrix.zeros(2, 2)
        matrix_c_inv_sympy[1, 1] = Integer(3)
        decay_mats_sympy = decaydata.DecayMatricesSympy(
            atomic_masses_sympy, decay_consts_sympy, matrix_c_sympy, matrix_c_inv_sympy
        )
        sympy_year_conv = Integer(3652422) / Integer(10000)

        dataset = decaydata.DecayData(
            dataset_name,
            bfs,
            float_year_conv,
            hldata,
            modes,
            nuclides,
            progeny,
            decay_mats,
            decay_mats_sympy,
            sympy_year_conv,
        )
        self.assertIsNotNone(dataset.sympy_data)
        self.assertIsNotNone(dataset.sympy_year_conv)

    def test_half_life(self) -> None:
        """
        Test DecayData half_life() method.
        """

        data = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        self.assertEqual(data.half_life("H-3"), 388781329.30560005)
        self.assertEqual(data.half_life("H-3", "y"), 12.32)
        self.assertEqual(data.half_life("Fm-257", "h"), 2412.0)
        self.assertEqual(data.half_life("Rn-222", "d"), 3.8235)

        self.assertEqual(data.half_life("H-3", "readable"), "12.32 y")
        self.assertEqual(data.half_life("Po-213", "readable"), "4.2 μs")
        self.assertEqual(data.half_life("Ra-219", "readable"), "10 ms")
        self.assertEqual(data.half_life("Rn-215", "readable"), "2.30 μs")
        self.assertEqual(data.half_life("U-238", "readable"), "4.468 By")

    def test_branching_fraction(self) -> None:
        """
        Test DecayData branching_fraction() method.
        """

        data = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        self.assertEqual(data.branching_fraction("K-40", "Ca-40"), 0.8914)
        self.assertEqual(data.branching_fraction("K-40", "H-3"), 0.0)
        self.assertEqual(data.branching_fraction("Cu-64", "Ni-64"), 0.61)

    def test_decay_mode(self) -> None:
        """
        Test DecayData decay_mode() method.
        """

        data = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        self.assertEqual(data.decay_mode("K-40", "Ca-40"), "\u03b2-")
        self.assertEqual(data.decay_mode("K-40", "H-3"), "")
        self.assertEqual(data.decay_mode("Cu-64", "Ni-64"), "\u03b2+ & EC")

    def test_decaydata___eq__(self) -> None:
        """
        Test DecayData equality.
        """

        data1 = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        data2 = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        self.assertEqual(data1, data2)

        self.assertFalse(data1 == "random object")

    def test_decaydata___ne__(self) -> None:
        """
        Test DecayData inequality.
        """

        data1 = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        data2 = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        data2.dataset_name = "icrp07"
        self.assertNotEqual(data1, data2)

        self.assertTrue(data1 != "random object")

    def test_decaydata___repr__(self) -> None:
        """
        Test DecayData __repr__ strings.
        """

        data = decaydata.load_dataset("icrp107_ame2020_nubase2020")
        self.assertEqual(
            repr(data),
            "Decay dataset: icrp107_ame2020_nubase2020, contains SymPy data: False",
        )

        data = decaydata.load_dataset("icrp107_ame2020_nubase2020", load_sympy=True)
        self.assertEqual(
            repr(data),
            "Decay dataset: icrp107_ame2020_nubase2020, contains SymPy data: True",
        )


class TestFileIOFunctions(unittest.TestCase):
    """
    Unit tests for the decaydata.py file I/O and dataset loading functions.
    """

    def test_load_dataset(self) -> None:
        """
        Test load_dataset() function.
        """

        # pylint: disable=too-many-statements

        # check instantiation from sub-package
        data = decaydata.load_dataset("icrp107_ame2020_nubase2020", load_sympy=False)
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[0][2], "100.5 d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.hldata[-1][2], "stable")
        self.assertEqual(data.hldata[1351][2], "stable")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "H-1")
        self.assertEqual(data.nuclides[1351], "Ni-58")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["H-1"], 1511)
        self.assertEqual(data.nuclide_dict["Ni-58"], 1351)
        self.assertEqual(data.progeny[0][0], "Cf-253")
        self.assertEqual(data.bfs[0][0], 0.9979)
        self.assertEqual(data.modes[0][0], "\u03b1")
        self.assertEqual(data.progeny[0][1], "SF")
        self.assertEqual(data.bfs[0][1], 0.0021)
        self.assertEqual(data.modes[0][1], "SF")
        self.assertEqual(data.progeny[-1], [])
        self.assertEqual(data.bfs[-1], [])
        self.assertEqual(data.modes[-1], [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        with self.assertRaises(ValueError):
            assert data.sympy_data
        with self.assertRaises(ValueError):
            assert data.sympy_year_conv

        # check instantiation with supplied dataset path
        data = decaydata.load_dataset(
            "icrp107_ame2020_nubase2020_2",
            icrp107_ame2020_nubase2020.__path__[0],
            load_sympy=False,
        )
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020_2")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[0][2], "100.5 d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.hldata[-1][2], "stable")
        self.assertEqual(data.hldata[1351][2], "stable")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "H-1")
        self.assertEqual(data.nuclides[1351], "Ni-58")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["H-1"], 1511)
        self.assertEqual(data.nuclide_dict["Ni-58"], 1351)
        self.assertEqual(data.progeny[0][0], "Cf-253")
        self.assertEqual(data.bfs[0][0], 0.9979)
        self.assertEqual(data.modes[0][0], "\u03b1")
        self.assertEqual(data.progeny[0][1], "SF")
        self.assertEqual(data.bfs[0][1], 0.0021)
        self.assertEqual(data.modes[0][1], "SF")
        self.assertEqual(data.progeny[-1], [])
        self.assertEqual(data.bfs[-1], [])
        self.assertEqual(data.modes[-1], [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        with self.assertRaises(ValueError):
            assert data.sympy_data
        with self.assertRaises(ValueError):
            assert data.sympy_year_conv

        # check instantiation from sub-package with SymPy data
        data = decaydata.load_dataset("icrp107_ame2020_nubase2020", load_sympy=True)
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[0][2], "100.5 d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.hldata[-1][2], "stable")
        self.assertEqual(data.hldata[1351][2], "stable")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "H-1")
        self.assertEqual(data.nuclides[1351], "Ni-58")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["H-1"], 1511)
        self.assertEqual(data.nuclide_dict["Ni-58"], 1351)
        self.assertEqual(data.progeny[0][0], "Cf-253")
        self.assertEqual(data.bfs[0][0], 0.9979)
        self.assertEqual(data.modes[0][0], "\u03b1")
        self.assertEqual(data.progeny[0][1], "SF")
        self.assertEqual(data.bfs[0][1], 0.0021)
        self.assertEqual(data.modes[0][1], "SF")
        self.assertEqual(data.progeny[-1], [])
        self.assertEqual(data.bfs[-1], [])
        self.assertEqual(data.modes[-1], [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        self.assertEqual(
            data.sympy_data.decay_consts[0],
            log(2) * Integer(10) / (Integer(1005) * Integer(24) * Integer(3600)),
        )
        self.assertEqual(data.sympy_data.ln2, log(2))
        self.assertEqual(data.sympy_data.matrix_c.shape, (1512, 1512))
        self.assertEqual(data.sympy_data.matrix_c[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_c_inv.shape, (1512, 1512))
        self.assertEqual(data.sympy_data.matrix_c_inv[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_e.shape, (1512, 1512))
        self.assertEqual(data.sympy_data.matrix_e[0, 0], Integer(0))
        self.assertEqual(data.sympy_data.vector_n0[0], Integer(0))
        self.assertEqual(data.sympy_year_conv, Integer(3652422) / Integer(10000))

        # check instantiation with supplied dataset path with SymPy data
        data = decaydata.load_dataset(
            "icrp107_ame2020_nubase2020_2",
            icrp107_ame2020_nubase2020.__path__[0],
            load_sympy=True,
        )
        self.assertEqual(data.dataset_name, "icrp107_ame2020_nubase2020_2")
        self.assertEqual(data.hldata[0][0], 100.5)
        self.assertEqual(data.hldata[0][1], "d")
        self.assertEqual(data.hldata[0][2], "100.5 d")
        self.assertEqual(data.hldata[-1][0], np.inf)
        self.assertEqual(data.hldata[-1][1], "s")
        self.assertEqual(data.hldata[-1][2], "stable")
        self.assertEqual(data.hldata[1351][2], "stable")
        self.assertEqual(data.nuclides[0], "Fm-257")
        self.assertEqual(data.nuclides[-1], "H-1")
        self.assertEqual(data.nuclides[1351], "Ni-58")
        self.assertEqual(data.nuclide_dict["Fm-257"], 0)
        self.assertEqual(data.nuclide_dict["H-1"], 1511)
        self.assertEqual(data.nuclide_dict["Ni-58"], 1351)
        self.assertEqual(data.progeny[0][0], "Cf-253")
        self.assertEqual(data.bfs[0][0], 0.9979)
        self.assertEqual(data.modes[0][0], "\u03b1")
        self.assertEqual(data.progeny[0][1], "SF")
        self.assertEqual(data.bfs[0][1], 0.0021)
        self.assertEqual(data.modes[0][1], "SF")
        self.assertEqual(data.progeny[-1], [])
        self.assertEqual(data.bfs[-1], [])
        self.assertEqual(data.modes[-1], [])
        self.assertEqual(
            data.scipy_data.decay_consts[0], np.log(2) / (100.5 * 24 * 60 * 60)
        )
        self.assertEqual(data.scipy_data.ln2, np.log(2))
        self.assertEqual(data.scipy_data.matrix_c.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_c_inv.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_c_inv[0, 0], 1.0)
        self.assertEqual(data.scipy_data.matrix_e.shape, (1512, 1512))
        self.assertEqual(data.scipy_data.matrix_e[0, 0], 0.0)
        self.assertEqual(data.scipy_data.vector_n0[0], 0.0)
        self.assertEqual(
            data.sympy_data.decay_consts[0],
            log(2) * Integer(10) / (Integer(1005) * Integer(24) * Integer(3600)),
        )
        self.assertEqual(data.sympy_data.ln2, log(2))
        self.assertEqual(data.sympy_data.matrix_c.shape, (1512, 1512))
        self.assertEqual(data.sympy_data.matrix_c[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_c_inv.shape, (1512, 1512))
        self.assertEqual(data.sympy_data.matrix_c_inv[0, 0], Integer(1))
        self.assertEqual(data.sympy_data.matrix_e.shape, (1512, 1512))
        self.assertEqual(data.sympy_data.matrix_e[0, 0], Integer(0))
        self.assertEqual(data.sympy_data.vector_n0[0], Integer(0))
        self.assertEqual(data.sympy_year_conv, Integer(3652422) / Integer(10000))


if __name__ == "__main__":
    unittest.main()



"""
The plots module defines functions used for creating decay chain diagrams via the Nuclide
class ``plot()`` method, and activity decay graphs via the Inventory class ``plot()`` method.

"""

from typing import List, Optional, Set, Tuple

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

# pylint: disable=too-many-arguments, too-many-locals


def _parse_nuclide_label(nuclide: str) -> str:
    """
    Format a nuclide string to mass number, meta-stable state character in
    superscript, then element symbol. Output is used on node labels in decay
    chain plots.

    Parameters
    ----------
    nuclide : str
        Nuclide string in element-mass format.

    Returns
    -------
    str
        Parsed string for node label in ^{mass}element format.

    """

    if nuclide == "SF":
        return "various"

    nuclide_conversion = {
        "0": "\N{SUPERSCRIPT ZERO}",
        "1": "\N{SUPERSCRIPT ONE}",
        "2": "\N{SUPERSCRIPT TWO}",
        "3": "\N{SUPERSCRIPT THREE}",
        "4": "\N{SUPERSCRIPT FOUR}",
        "5": "\N{SUPERSCRIPT FIVE}",
        "6": "\N{SUPERSCRIPT SIX}",
        "7": "\N{SUPERSCRIPT SEVEN}",
        "8": "\N{SUPERSCRIPT EIGHT}",
        "9": "\N{SUPERSCRIPT NINE}",
        "m": "\N{MODIFIER LETTER SMALL M}",
        "n": "\N{SUPERSCRIPT LATIN SMALL LETTER N}",
        "p": "\N{MODIFIER LETTER SMALL P}",
        "q": "\N{LATIN SMALL LETTER Q}",  # Unicode has no superscript q
        "r": "\N{MODIFIER LETTER SMALL R}",
        "x": "\N{MODIFIER LETTER SMALL X}",
    }

    element, isotope = nuclide.split("-")
    return "".join(map(lambda char: nuclide_conversion[char], list(isotope))) + element


def _parse_decay_mode_label(mode: str) -> str:
    """
    Format a decay mode string for edge label on decay chain plot.

    Parameters
    ----------
    mode : str
        Decay mode string.

    Returns
    -------
    str
        Formatted decay mode string for use in an edge label.

    """

    mode_conversion = {
        "α": "\N{GREEK SMALL LETTER ALPHA}",
        "β": "\N{GREEK SMALL LETTER BETA}",
        "ε": "\N{GREEK SMALL LETTER EPSILON}",
        "+": "\N{SUPERSCRIPT PLUS SIGN}",
        "-": "\N{SUPERSCRIPT MINUS}",
        "12C": "\N{SUPERSCRIPT ONE}\N{SUPERSCRIPT TWO}C",
        "14C": "\N{SUPERSCRIPT ONE}\N{SUPERSCRIPT FOUR}C",
        "20O": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT ZERO}O",
        "23F": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT THREE}F",
        "22Ne": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT TWO}Ne",
        "24Ne": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT FOUR}Ne",
        "25Ne": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT FIVE}Ne",
        "26Ne": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT SIX}Ne",
        "28Mg": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT EIGHT}Mg",
        "29Mg": "\N{SUPERSCRIPT TWO}\N{SUPERSCRIPT NINE}Mg",
        "30Mg": "\N{SUPERSCRIPT THREE}\N{SUPERSCRIPT ZERO}Mg",
        "32Si": "\N{SUPERSCRIPT THREE}\N{SUPERSCRIPT TWO}Si",
        "34Si": "\N{SUPERSCRIPT THREE}\N{SUPERSCRIPT FOUR}Si",
    }

    for unformatted, formatted in mode_conversion.items():
        mode = mode.replace(unformatted, formatted)
    return mode


def _check_fig_axes(  # type: ignore
    fig_in: Optional[matplotlib.figure.Figure],
    axes_in: Optional[matplotlib.axes.Axes],
    **kwargs,
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """
    Checks to see if user supplies Matplotlib Figure and/or Axes objects. Creates them where
    necessary.

    Parameters
    ----------
    fig_in : None or matplotlib.figure.Figure
        matplotlib figure object to use, or None creates one.
    axes_in : matplotlib.axes.Axes or None, optional
        matplotlib axes object to use, or None creates one.
    **kwargs
        All additional keyword arguments to supply to plt.subplots().

    Returns
    -------
    fig : matplotlib.figure.Figure
        matplotlib figure object used to plot decay chain.
    axes : matplotlib.axes.Axes
        matplotlib axes object used to plot decay chain.

    """

    if fig_in is None and axes_in is None:
        fig, axes = plt.subplots(**kwargs)
    elif fig_in is None:
        axes = axes_in
        fig = axes.get_figure()
    elif axes_in is None:
        fig = fig_in
        axes = fig.gca()
    else:
        fig = fig_in
        axes = axes_in

    return fig, axes


def decay_graph(  # type: ignore
    time_points: np.ndarray,
    ydata: np.ndarray,
    nuclides: List[str],
    xunits: str,
    ylabel: str,
    xscale: str,
    yscale: str,
    ylimits: List[float],
    display: Set[str],
    fig_in: Optional[matplotlib.figure.Figure],
    axes_in: Optional[matplotlib.axes.Axes],
    **kwargs,
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """
    Plots a decay graph showing the change in activity of an inventory over time. Creates
    matplotlib fig, axes objects if they are not supplied. Returns fig, axes tuple.

    Parameters
    ----------
    time_points : numpy.ndarray
        Time points for x-axis.
    ydata : numpy.ndarray
        y-axis data.
    nuclides : list
        List of the nuclides (string format is 'H-3', etc.).
    xunits : str
        Units for decay time axis.
    ylabel : str
        Units for the y-axis
    xscale : str
        The time axis scale type to apply ('linear' or 'log').
    yscale : str
        The y-axis scale type to apply ('linear' or 'log').
    ylimits : list
        Limits for the y-axis (list or numpy.ndarray with two elements).
    display : set of str
        Nuclides to display on the graph.
    fig_in : None or matplotlib.figure.Figure
        matplotlib figure object to use, or None creates one.
    axes_in : matplotlib.axes.Axes or None, optional
        matplotlib axes object to use, or None creates one.
    **kwargs
        All additional keyword arguments to supply to matplotlib plot().

    Returns
    -------
    fig : matplotlib.figure.Figure
        matplotlib figure object used to plot decay chain.
    axes : matplotlib.axes.Axes
        matplotlib axes object used to plot decay chain.

    """

    fig, axes = _check_fig_axes(fig_in, axes_in)

    for idx, label in enumerate(nuclides):
        if label in display:
            axes.plot(time_points, ydata[idx], label=label, **kwargs)
    axes.legend(loc="upper right")
    xlabel = f"Time ({xunits})"
    axes.set(
        xlabel=xlabel,
        ylabel=ylabel,
        xscale=xscale,
        yscale=yscale,
    )
    axes.set_ylim(ylimits)

    return fig, axes



"""
Unit tests for converts.py classes and methods.
"""

import unittest

from sympy import Integer, log

from radioactivedecay.converters import (
    AVOGADRO,
    QuantityConverterFloat,
    QuantityConverterSympy,
    UnitConverterFloat,
    UnitConverterSympy,
)


class TestConverters(unittest.TestCase):
    """
    Unit tests for the converters.py constants.
    """

    def test_avogadro(self) -> None:
        """
        Test module constants.
        """

        self.assertEqual(AVOGADRO, 6.02214076e23)


class TestUnitConverterFloat(unittest.TestCase):
    """
    Unit tests for the converters.py UnitConverterFloat class.
    """

    def test_class_attributes(self) -> None:
        """
        Test class attributes of UnitConverterFloat objects.
        """

        self.assertEqual(UnitConverterFloat.time_units["s"], 1.0)
        self.assertEqual(UnitConverterFloat.time_units["y"], 86400.0)
        self.assertEqual(UnitConverterFloat.activity_units["Bq"], 1.0)
        self.assertEqual(UnitConverterFloat.mass_units["g"], 1.0)
        self.assertEqual(UnitConverterFloat.moles_units["mol"], 1.0)

        self.assertEqual(
            UnitConverterFloat.time_unit_err_msg,
            'is not a valid time unit, e.g. "s", "m", "h", "d" or "y".',
        )
        self.assertEqual(
            UnitConverterFloat.activity_unit_err_msg,
            'is not a valid activitiy unit, e.g. "Bq", "kBq", "Ci"...',
        )
        self.assertEqual(
            UnitConverterFloat.mass_unit_err_msg,
            'is not a valid mass unit, e.g. "g", "kg", "mg"...',
        )
        self.assertEqual(
            UnitConverterFloat.moles_unit_err_msg,
            'is not a valid moles unit, e.g. "mol", "kmol", "mmol"...',
        )

    def test_time_unit_conv_seconds(self) -> None:
        """
        Test conversion between seconds and different time units.
        """

        year_conv = 365.2422

        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "s", year_conv), 1.0e0
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "ps", year_conv),
            1.0e12,
            places=(15 - 12),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "ns", year_conv),
            1.0e9,
            places=(15 - 9),
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "μs", year_conv), 1.0e6
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "us", year_conv), 1.0e6
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "ms", year_conv), 1.0e3
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "m", year_conv), 1.0 / 60.0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "h", year_conv),
            1.0 / (60.0**2),
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "d", year_conv),
            1.0 / (60.0**2 * 24.0),
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "y", year_conv),
            1.0 / (60.0**2 * 24.0 * year_conv),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "ps", "s", year_conv),
            1.0e-12,
            places=(12 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "ns", "s", year_conv),
            1.0e-9,
            places=(9 + 15),
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "μs", "s", year_conv), 1.0e-6
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "us", "s", year_conv), 1.0e-6
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "ms", "s", year_conv), 1.0e-3
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "m", "s", year_conv), 60.0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "h", "s", year_conv), (60.0**2)
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "d", "s", year_conv),
            (60.0**2 * 24.0),
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "s", year_conv),
            (60.0**2 * 24.0 * year_conv),
        )

        # Catch some incorrect time units
        with self.assertRaises(ValueError):
            UnitConverterFloat.time_unit_conv(1.0, "ty", "y", year_conv)
        with self.assertRaises(ValueError):
            UnitConverterFloat.time_unit_conv(1.0, "y", "ty", year_conv)
        with self.assertRaises(ValueError):
            UnitConverterFloat.time_unit_conv(1.0, "ty", 1.0, year_conv)

    def test_time_unit_conv_spelling_variations(self) -> None:
        """
        Test spelling variations of different time units.
        """

        year_conv = 365.2422

        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "sec", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "second", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "s", "seconds", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "h", "hr", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "h", "hour", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "h", "hours", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "d", "day", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "d", "days", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "yr", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "year", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "years", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "sec", "s", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "second", "s", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "seconds", "s", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "hr", "h", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "hour", "h", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "hours", "h", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "day", "d", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "days", "d", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "yr", "y", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "year", "y", year_conv), 1.0e0
        )
        self.assertEqual(
            UnitConverterFloat.time_unit_conv(1.0, "years", "y", year_conv), 1.0e0
        )

    def test_time_unit_conv_year_prefixes(self) -> None:
        """
        Test conversions between different year prefixes.
        """

        year_conv = 365.2422

        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "ky", year_conv),
            1.0e-3,
            places=(3 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "My", year_conv),
            1.0e-6,
            places=(6 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "By", year_conv),
            1.0e-9,
            places=(9 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "Gy", year_conv),
            1.0e-9,
            places=(9 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "Ty", year_conv),
            1.0e-12,
            places=(12 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "y", "Py", year_conv),
            1.0e-15,
            places=(15 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "ky", "y", year_conv),
            1.0e3,
            places=(15 - 3),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "My", "y", year_conv),
            1.0e6,
            places=(15 - 6),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "Gy", "y", year_conv),
            1.0e9,
            places=(15 - 9),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "Ty", "y", year_conv),
            1.0e12,
            places=(15 - 12),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.time_unit_conv(1.0, "Py", "y", year_conv),
            1.0e15,
            places=(15 - 15),
        )

    def test_activity_unit_conv(self) -> None:
        """
        Test conversions between activity units.
        """

        self.assertEqual(UnitConverterFloat.activity_unit_conv(1.0, "Bq", "Bq"), 1.0e0)
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "pBq"),
            1.0e12,
            places=(15 - 12),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "nBq"),
            1.0e9,
            places=(15 - 9),
        )
        self.assertEqual(UnitConverterFloat.activity_unit_conv(1.0, "Bq", "μBq"), 1.0e6)
        self.assertEqual(UnitConverterFloat.activity_unit_conv(1.0, "Bq", "uBq"), 1.0e6)
        self.assertEqual(UnitConverterFloat.activity_unit_conv(1.0, "Bq", "mBq"), 1.0e3)
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "kBq"),
            1.0e-3,
            places=(3 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "MBq"),
            1.0e-6,
            places=(6 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "GBq"),
            1.0e-9,
            places=(9 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "TBq"),
            1.0e-12,
            places=(12 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "PBq"),
            1.0e-15,
            places=(15 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "EBq"),
            1.0e-18,
            places=(18 + 15),
        )

        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "pCi"),
            1.0e12 / 3.7e10,
            places=(15 - 12),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "nCi"),
            1.0e9 / 3.7e10,
            places=(15 - 9),
        )
        self.assertEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "μCi"), 1.0e6 / 3.7e10
        )
        self.assertEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "uCi"), 1.0e6 / 3.7e10
        )
        self.assertEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "mCi"), 1.0e3 / 3.7e10
        )
        self.assertEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "Ci"), 1.0 / 3.7e10
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "kCi"),
            1.0e-3 / 3.7e10,
            places=(3 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "MCi"),
            1.0e-6 / 3.7e10,
            places=(6 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "GCi"),
            1.0e-9 / 3.7e10,
            places=(9 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "TCi"),
            1.0e-12 / 3.7e10,
            places=(12 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "PCi"),
            1.0e-15 / 3.7e10,
            places=(15 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "ECi"),
            1.0e-18 / 3.7e10,
            places=(18 + 15),
        )

        self.assertEqual(UnitConverterFloat.activity_unit_conv(1.0, "Bq", "dpm"), 60.0)

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            UnitConverterFloat.activity_unit_conv(1.0, "tBq", "Bq")
        with self.assertRaises(ValueError):
            UnitConverterFloat.activity_unit_conv(1.0, "Bq", "tBq")
        with self.assertRaises(ValueError):
            UnitConverterFloat.activity_unit_conv(1.0, "tBq", 1.0)

    def test_mass_unit_conv(self) -> None:
        """
        Test conversions between mass units.
        """

        self.assertEqual(UnitConverterFloat.mass_unit_conv(1.0, "g", "g"), 1.0e0)
        self.assertAlmostEqual(
            UnitConverterFloat.mass_unit_conv(1.0, "g", "pg"), 1.0e12, places=(15 - 12)
        )
        self.assertAlmostEqual(
            UnitConverterFloat.mass_unit_conv(1.0, "g", "ng"), 1.0e9, places=(15 - 9)
        )
        self.assertEqual(UnitConverterFloat.mass_unit_conv(1.0, "g", "μg"), 1.0e6)
        self.assertEqual(UnitConverterFloat.mass_unit_conv(1.0, "g", "ug"), 1.0e6)
        self.assertEqual(UnitConverterFloat.mass_unit_conv(1.0, "g", "mg"), 1.0e3)
        self.assertAlmostEqual(
            UnitConverterFloat.mass_unit_conv(1.0, "g", "kg"), 1.0e-3, places=(3 + 15)
        )
        self.assertAlmostEqual(
            UnitConverterFloat.mass_unit_conv(1.0, "g", "Mg"), 1.0e-6, places=(6 + 15)
        )
        self.assertAlmostEqual(
            UnitConverterFloat.mass_unit_conv(1.0, "g", "t"), 1.0e-6, places=(6 + 15)
        )
        self.assertAlmostEqual(
            UnitConverterFloat.mass_unit_conv(1.0, "g", "ton"), 1.0e-6, places=(6 + 15)
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            UnitConverterFloat.mass_unit_conv(1.0, "tg", "g")
        with self.assertRaises(ValueError):
            UnitConverterFloat.mass_unit_conv(1.0, "g", "tg")
        with self.assertRaises(ValueError):
            UnitConverterFloat.mass_unit_conv(1.0, "tg", 1.0)

    def test_moles_unit_conv(self) -> None:
        """
        Test conversions between moles orders of magnitude.
        """

        self.assertEqual(UnitConverterFloat.moles_unit_conv(1.0, "mol", "mol"), 1.0e0)
        self.assertAlmostEqual(
            UnitConverterFloat.moles_unit_conv(1.0, "mol", "pmol"),
            1.0e12,
            places=(15 - 12),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.moles_unit_conv(1.0, "mol", "nmol"),
            1.0e9,
            places=(15 - 9),
        )
        self.assertEqual(UnitConverterFloat.moles_unit_conv(1.0, "mol", "μmol"), 1.0e6)
        self.assertEqual(UnitConverterFloat.moles_unit_conv(1.0, "mol", "umol"), 1.0e6)
        self.assertEqual(UnitConverterFloat.moles_unit_conv(1.0, "mol", "mmol"), 1.0e3)
        self.assertAlmostEqual(
            UnitConverterFloat.moles_unit_conv(1.0, "mol", "kmol"),
            1.0e-3,
            places=(3 + 15),
        )
        self.assertAlmostEqual(
            UnitConverterFloat.moles_unit_conv(1.0, "mol", "Mmol"),
            1.0e-6,
            places=(6 + 15),
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            UnitConverterFloat.moles_unit_conv(1.0, "tmol", "mol")
        with self.assertRaises(ValueError):
            UnitConverterFloat.moles_unit_conv(1.0, "mol", "tmol")
        with self.assertRaises(ValueError):
            UnitConverterFloat.moles_unit_conv(1.0, "tmol", 1.0)

    def test___repr__(self) -> None:
        """
        Test UnitConverterFloat __repr__ strings.
        """

        self.assertEqual(
            repr(UnitConverterFloat()),
            "UnitConverterFloat using double-precision floats.",
        )


class TestUnitConverterSympy(unittest.TestCase):
    """
    Unit tests for the converters.py UnitConverterSympy class.
    """

    def test_class_attributes(self) -> None:
        """
        Test class attributes of UnitConverterSympy objects.
        """

        self.assertEqual(UnitConverterSympy.time_units["s"], Integer(1))
        self.assertEqual(UnitConverterSympy.time_units["y"], Integer(86400))
        self.assertEqual(UnitConverterSympy.activity_units["Bq"], Integer(1))
        self.assertEqual(UnitConverterSympy.mass_units["g"], Integer(1))
        self.assertEqual(UnitConverterSympy.moles_units["mol"], Integer(1))

    def test_time_unit_conv(self) -> None:
        """
        Test of the SymPy version of time_unit_conv().
        """

        year_conv = Integer(3652422) / 10000

        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "ps", "ns", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "ns", "us", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "μs", "ms", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "us", "ms", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "ms", "s", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "s", "m", year_conv),
            1 / Integer(60),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "m", "h", year_conv),
            1 / Integer(60),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "h", "d", year_conv),
            1 / Integer(24),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "d", "y", year_conv),
            10000 / Integer(3652422),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "y", "ky", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "ky", "My", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "My", "By", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "My", "Gy", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "By", "Ty", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "Gy", "Ty", year_conv),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "Ty", "Py", year_conv),
            1 / Integer(1000),
        )

        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "ns", "ps", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "μs", "ns", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "us", "ns", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "ms", "us", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "s", "ms", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "m", "s", year_conv),
            Integer(60),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "h", "m", year_conv),
            Integer(60),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "d", "h", year_conv),
            Integer(24),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "y", "d", year_conv),
            Integer(3652422) / 10000,
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "ky", "y", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "My", "ky", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "By", "My", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "Gy", "My", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "Ty", "By", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "Ty", "Gy", year_conv),
            Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.time_unit_conv(Integer(1), "Py", "Ty", year_conv),
            Integer(1000),
        )

        # Catch some incorrect time units
        with self.assertRaises(ValueError):
            UnitConverterSympy.time_unit_conv(Integer(1), "ty", "y", year_conv)
        with self.assertRaises(ValueError):
            UnitConverterSympy.time_unit_conv(Integer(1), "y", "ty", year_conv)
        with self.assertRaises(ValueError):
            UnitConverterSympy.time_unit_conv(Integer(1), "ty", Integer(1), year_conv)

    def test_activity_unit_conv(self) -> None:
        """
        Test of the SymPy version of activity_unit_conv().
        """

        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "pBq", "nBq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "nBq", "μBq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "μBq", "uBq"), Integer(1)
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "uBq", "mBq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "mBq", "Bq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "Bq", "kBq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "kBq", "MBq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "MBq", "GBq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "GBq", "TBq"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "TBq", "PBq"),
            1 / Integer(1000),
        )

        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "pBq", "pCi"),
            1 / Integer(37000000000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "pCi", "nCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "nCi", "μCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "μCi", "uCi"), Integer(1)
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "uCi", "mCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "mCi", "Ci"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "Ci", "kCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "kCi", "MCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "MCi", "GCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "GCi", "TCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "TCi", "PCi"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "PCi", "ECi"),
            1 / Integer(1000),
        )

        self.assertEqual(
            UnitConverterSympy.activity_unit_conv(Integer(1), "Bq", "dpm"),
            Integer(60),
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            UnitConverterSympy.activity_unit_conv(Integer(1), "tBq", "Bq")
        with self.assertRaises(ValueError):
            UnitConverterSympy.activity_unit_conv(Integer(1), "Bq", "tBq")
        with self.assertRaises(ValueError):
            UnitConverterSympy.activity_unit_conv(Integer(1), "tBq", Integer(1))

    def test_mass_unit_conv(self) -> None:
        """
        Test of the SymPy version of mass_unit_conv().
        """

        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "pg", "ng"), 1 / Integer(1000)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "ng", "μg"), 1 / Integer(1000)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "μg", "ug"), Integer(1)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "ug", "mg"), 1 / Integer(1000)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "mg", "g"), 1 / Integer(1000)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "g", "kg"), 1 / Integer(1000)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "kg", "Mg"), 1 / Integer(1000)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "Mg", "t"), Integer(1)
        )
        self.assertEqual(
            UnitConverterSympy.mass_unit_conv(Integer(1), "Mg", "ton"), Integer(1)
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            UnitConverterSympy.mass_unit_conv(Integer(1), "tg", "g")
        with self.assertRaises(ValueError):
            UnitConverterSympy.mass_unit_conv(Integer(1), "g", "tg")
        with self.assertRaises(ValueError):
            UnitConverterSympy.mass_unit_conv(Integer(1), "tg", Integer(1))

    def test_moles_unit_conv(self) -> None:
        """
        Test of the SymPy version of moles_unit_conv().
        """

        self.assertEqual(
            UnitConverterSympy.moles_unit_conv(Integer(1), "pmol", "nmol"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.moles_unit_conv(Integer(1), "nmol", "μmol"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.moles_unit_conv(Integer(1), "μmol", "umol"), Integer(1)
        )
        self.assertEqual(
            UnitConverterSympy.moles_unit_conv(Integer(1), "umol", "mmol"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.moles_unit_conv(Integer(1), "mmol", "mol"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.moles_unit_conv(Integer(1), "mol", "kmol"),
            1 / Integer(1000),
        )
        self.assertEqual(
            UnitConverterSympy.moles_unit_conv(Integer(1), "kmol", "Mmol"),
            1 / Integer(1000),
        )

        # Catch some incorrect activity units
        with self.assertRaises(ValueError):
            UnitConverterSympy.moles_unit_conv(Integer(1), "tmol", "mol")
        with self.assertRaises(ValueError):
            UnitConverterSympy.moles_unit_conv(Integer(1), "mol", "tmol")
        with self.assertRaises(ValueError):
            UnitConverterSympy.moles_unit_conv(Integer(1), "tmol", Integer(1))

    def test___repr__(self) -> None:
        """
        Test UnitConverterSympy __repr__ strings.
        """

        self.assertEqual(
            repr(UnitConverterSympy()),
            "UnitConverterSympy using SymPy arbitrary precision calculations.",
        )


class TestQuantityConverterFloat(unittest.TestCase):
    """
    Unit tests for the converters.py QuantityConverterFloat class.
    """

    def test_class_attribute(self) -> None:
        """
        Test class attribute of QuantityConverterFloat.
        """

        self.assertEqual(QuantityConverterFloat.avogadro, 6.02214076e23)

    def test_activity_to_number(self) -> None:
        """
        Test the conversion of activity in Bq to number of atoms.
        """

        self.assertEqual(
            QuantityConverterFloat.activity_to_number(1.0, 1.7828715741004621e-09),
            560892895.7794082,
        )

    def test_mass_to_number(self) -> None:
        """
        Test the conversion of mass in grams to number of atoms.
        """

        self.assertEqual(
            QuantityConverterFloat.mass_to_number(1.0, 3.01604928132),
            1.996698395247825e23,
        )
        self.assertEqual(
            QuantityConverterFloat.mass_to_number(1.0, 3.01602932197),
            1.9967116089131645e23,
        )

    def test_moles_to_number(self) -> None:
        """
        Test the conversion of number of moles to number of atoms.
        """

        self.assertEqual(QuantityConverterFloat.moles_to_number(1.0), 6.02214076e23)
        self.assertEqual(QuantityConverterFloat.moles_to_number(0.0), 0.0)

    def test_number_to_activity(self) -> None:
        """
        Test the conversion of number of atoms to activity in Bq.
        """

        self.assertEqual(
            QuantityConverterFloat.number_to_activity(
                560892895.7794082, 1.7828715741004621e-09
            ),
            1.0,
        )
        self.assertEqual(QuantityConverterFloat.number_to_activity(1.0, 0.0), 0.0)
        self.assertEqual(QuantityConverterFloat.number_to_activity(0.0, 0.0), 0.0)

    def test_number_to_mass(self) -> None:
        """
        Test the conversion of number of atoms to mass in grams.
        """

        self.assertEqual(
            QuantityConverterFloat.number_to_mass(1.996698395247825e23, 3.01604928132),
            1.0,
        )
        self.assertEqual(
            QuantityConverterFloat.number_to_mass(1.9967116089131645e23, 3.01602932197),
            1.0,
        )

    def test_number_to_moles(self) -> None:
        """
        Test the conversion of number of atoms to number of moles.
        """

        self.assertEqual(QuantityConverterFloat.number_to_moles(6.02214076e23), 1.0)
        self.assertEqual(QuantityConverterFloat.number_to_moles(0.0), 0.0)

    def test___repr__(self) -> None:
        """
        Test QuantityConverterFloat __repr__ strings.
        """

        self.assertEqual(
            repr(QuantityConverterFloat()),
            "QuantityConverterFloat using double-precision floats.",
        )


class TestQuantityConverterSympy(unittest.TestCase):
    """
    Unit tests for the converters.py QuantityConverterSympy class.
    """

    def test_class_attribute(self) -> None:
        """
        Test class attribute of QuantityConverterSympy class.
        """

        self.assertEqual(
            QuantityConverterSympy.avogadro, Integer(602214076000000000000000)
        )

    def test_activity_to_number(self) -> None:
        """
        Test the conversion of activity in Bq to number of atoms.
        """

        self.assertEqual(
            QuantityConverterSympy.activity_to_number(
                Integer(1), (Integer(625) * log(2)) / Integer(242988330816)
            ),
            Integer(242988330816) / (Integer(625) * log(2)),
        )

    def test_mass_to_number(self) -> None:
        """
        Test the conversion of mass in grams to number of atoms.
        """

        self.assertEqual(
            QuantityConverterSympy.mass_to_number(
                Integer(1),
                Integer("15055351900000000000000000000000000") / Integer(75401232033),
            ),
            Integer(75401232033)
            / Integer("15055351900000000000000000000000000")
            * Integer("602214076000000000000000"),
        )
        self.assertEqual(
            QuantityConverterSympy.mass_to_number(
                Integer(1),
                Integer("60221407600000000000000000000000000") / Integer(301602932197),
            ),
            Integer(301602932197)
            / Integer("60221407600000000000000000000000000")
            * Integer("602214076000000000000000"),
        )

    def test_moles_to_number(self) -> None:
        """
        Test the conversion of number of moles to number of atoms.
        """

        self.assertEqual(
            QuantityConverterSympy.moles_to_number(Integer(1)),
            Integer("602214076000000000000000"),
        )
        self.assertEqual(QuantityConverterSympy.moles_to_number(Integer(0)), Integer(0))

    def test_number_to_activity(self) -> None:
        """
        Test the conversion of number of atoms to activity in Bq.
        """

        self.assertEqual(
            QuantityConverterSympy.number_to_activity(
                Integer(1), (Integer(625) * log(2)) / Integer(242988330816)
            ),
            Integer(625) * log(2) / Integer(242988330816),
        )
        self.assertEqual(
            QuantityConverterSympy.number_to_activity(
                Integer(0), (Integer(625) * log(2)) / Integer(242988330816)
            ),
            Integer(0),
        )
        self.assertEqual(
            QuantityConverterSympy.number_to_activity(Integer(0), Integer(0)),
            Integer(0),
        )

    def test_number_to_mass(self) -> None:
        """
        Test the conversion of number of atoms to mass in grams.
        """

        self.assertEqual(
            QuantityConverterSympy.number_to_mass(
                Integer(1),
                Integer("15055351900000000000000000000000000") / Integer(75401232033),
            ),
            Integer("15055351900000000000000000000000000")
            / Integer(75401232033)
            / Integer("602214076000000000000000"),
        )
        self.assertEqual(
            QuantityConverterSympy.number_to_mass(
                Integer(1),
                Integer("60221407600000000000000000000000000") / Integer(301602932197),
            ),
            Integer("60221407600000000000000000000000000")
            / Integer(301602932197)
            / Integer("602214076000000000000000"),
        )

    def test_number_to_moles(self) -> None:
        """
        Test the conversion of number of atoms to number of moles.
        """

        self.assertEqual(
            QuantityConverterSympy.number_to_moles(Integer("602214076000000000000000")),
            Integer(1),
        )
        self.assertEqual(QuantityConverterSympy.number_to_moles(Integer(0)), Integer(0))

    def test___repr__(self) -> None:
        """
        Test QuantityConveterSympy __repr__ strings.
        """

        self.assertEqual(
            repr(QuantityConverterSympy()),
            "QuantityConverterSympy using SymPy arbitrary precision calculations.",
        )


if __name__ == "__main__":
    unittest.main()
